@@ -101,21 +101,6 @@ using std::make_pair;
 
 namespace {
 
-template<typename T>
-inline T get_norm(T* v, int f) {
-  T sq_norm = 0;
-  for (int z = 0; z < f; z++)
-    sq_norm += v[z] * v[z];
-  return sqrt(sq_norm);
-}
-
-template<typename T>
-inline void normalize(T* v, int f) {
-  T norm = get_norm(v, f);
-  for (int z = 0; z < f; z++)
-    v[z] /= norm;
-}
-
 template<typename T>
 inline T dot(const T* x, const T* y, int f) {
   T s = 0;
@@ -128,21 +113,13 @@ inline T dot(const T* x, const T* y, int f) {
 }
 
 template<typename T>
-inline void dot3(const T* x, const T* y, int f, T* xx, T* yy, T* xy) {
-  // This function computes x*y as well as x*x and y*y at the same time
-  *xx = 0;
-  *yy = 0;
-  *xy = 0;
-  for (int z = 0; z < f; z++) {
-    *xx += (*x) * (*x);
-    *yy += (*y) * (*y);
-    *xy += (*x) * (*y);
-    x++;
-    y++;
-  }
+inline T manhattan_distance(const T* x, const T* y, int f) {
+  T d = 0.0;
+  for (int i = 0; i < f; i++)
+    d += fabs(x[i] - y[i]);
+  return d;
 }
 
-
 #ifdef USE_AVX
 // Horizontal single sum of 256bit vector.
 inline float hsum256_ps_avx(__m256 v) {
@@ -175,78 +152,50 @@ inline float dot<float>(const float* x, const float *y, int f) {
 }
 
 template<>
-inline void dot3<float>(const float* x, const float *y, int f, float* xx, float* yy, float* xy) {
-  *xx = 0;
-  *yy = 0;
-  *xy = 0;
+inline float manhattan_distance<float>(const float* x, const float* y, int f) {
+  float result = 0;
+  int i = f;
   if (f > 7) {
-    __m256 xx_vec = _mm256_setzero_ps(), yy_vec = _mm256_setzero_ps(), xy_vec = _mm256_setzero_ps();
-    for (; f > 7; f -= 8) {
-      const __m256 a = _mm256_loadu_ps(x);
-      const __m256 b = _mm256_loadu_ps(y);
-      xx_vec = _mm256_add_ps(xx_vec, _mm256_mul_ps(a, a));
-      yy_vec = _mm256_add_ps(yy_vec, _mm256_mul_ps(b, b));
-      xy_vec = _mm256_add_ps(xy_vec, _mm256_mul_ps(a, b));
+    __m256 manhattan = _mm256_setzero_ps();
+    __m256 minus_zero = _mm256_set1_ps(-0.0f);
+    for (; i > 7; i -= 8) {
+      const __m256 x_minus_y = _mm256_sub_ps(_mm256_loadu_ps(x), _mm256_loadu_ps(y));
+      const __m256 distance = _mm256_andnot_ps(minus_zero, x_minus_y); // Absolute value of x_minus_y (forces sign bit to zero)
+      manhattan = _mm256_add_ps(manhattan, distance);
       x += 8;
       y += 8;
     }
-    // Sum all floats in xx, yy and xy register.
-    *xx = hsum256_ps_avx(xx_vec);
-    *yy = hsum256_ps_avx(yy_vec);
-    *xy = hsum256_ps_avx(xy_vec);
+    // Sum all floats in manhattan register.
+    result = hsum256_ps_avx(manhattan);
   }
   // Don't forget the remaining values.
-  for (; f > 0; f--) {
-    *xx += (*x) * (*x);
-    *yy += (*y) * (*y);
-    *xy += (*x) * (*y);
+  for (; i > 0; i--) {
+    result += fabsf(*x - *y);
     x++;
     y++;
   }
+  return result;
 }
 
-template<>
-inline float get_norm<float>(float *v, int f) {
-  float sq_norm = 0;
-  int i = f;
-  if (f > 7) {
-    __m256 norm = _mm256_setzero_ps();
-    for (; i > 7; i -= 8) {
-      const __m256 v_v = _mm256_loadu_ps(v);
-      norm = _mm256_add_ps(norm, _mm256_mul_ps(v_v, v_v));
-      v += 8;
-    }
-    // Sum all floats in norm register.
-    sq_norm = hsum256_ps_avx(norm);
-  }
-  // Don't forget the remaining values.
-  for (; i > 0; i--) {
-    sq_norm += *v * *v;
-    v++;
-  }
-  return sqrt(sq_norm);
-}
-
-template<>
-inline void normalize<float>(float *v, int f) {
-  float norm = get_norm(v, f);
-  __m256 v_norm = _mm256_set1_ps(norm);
-
-  int i = f;
-  for (; i > 7; i -= 8) {
-    _mm256_storeu_ps(v, _mm256_div_ps(_mm256_loadu_ps(v), v_norm));
-    v += 8;
-  }
-  // Don't forget the remaining values.
-  for (; i > 0; i--) {
-    *v /= norm;
-    v++;
-  }
-}
 #endif
 
+ 
+template<typename T>
+inline T get_norm(T* v, int f) {
+  return sqrt(dot(v, v, f));
+}
+
+template<typename T>
+inline void normalize(T* v, int f) {
+  T norm = get_norm(v, f);
+  if (norm > 0) {
+    for (int z = 0; z < f; z++)
+      v[z] /= norm;
+  }
+}
+
 template<typename T, typename Random, typename Distance, typename Node>
-inline void two_means(const vector<Node*>& nodes, int f, Random& random, bool cosine, T* iv, T* jv) {
+inline void two_means(const vector<Node*>& nodes, int f, Random& random, bool cosine, Node* p, Node* q) {
   /*
     This algorithm is a huge heuristic. Empirically it works really well, but I
     can't motivate it well. The basic idea is to keep two centroids and assign
@@ -259,23 +208,30 @@ inline void two_means(const vector<Node*>& nodes, int f, Random& random, bool co
   size_t i = random.index(count);
   size_t j = random.index(count-1);
   j += (j >= i); // ensure that i != j
-  memcpy(iv, nodes[i]->v, f * sizeof(T));
-  memcpy(jv, nodes[j]->v, f * sizeof(T));
-  if (cosine) { normalize(&iv[0], f); normalize(&jv[0], f); }
+  memcpy(p->v, nodes[i]->v, f * sizeof(T));
+  memcpy(q->v, nodes[j]->v, f * sizeof(T));
+  if (cosine) { normalize(p->v, f); normalize(q->v, f); }
+  Distance::init_node(p, f);
+  Distance::init_node(q, f);
 
   int ic = 1, jc = 1;
   for (int l = 0; l < iteration_steps; l++) {
     size_t k = random.index(count);
-    T di = ic * Distance::distance(&iv[0], nodes[k]->v, f),
-      dj = jc * Distance::distance(&jv[0], nodes[k]->v, f);
+    T di = ic * Distance::distance(p, nodes[k], f),
+      dj = jc * Distance::distance(q, nodes[k], f);
     T norm = cosine ? get_norm(nodes[k]->v, f) : 1.0;
+    if (!(norm > T(0))) {
+      continue;
+    }
     if (di < dj) {
       for (int z = 0; z < f; z++)
-	iv[z] = (iv[z] * ic + nodes[k]->v[z] / norm) / (ic + 1);
+	p->v[z] = (p->v[z] * ic + nodes[k]->v[z] / norm) / (ic + 1);
+      Distance::init_node(p, f);
       ic++;
     } else if (dj < di) {
       for (int z = 0; z < f; z++)
-	jv[z] = (jv[z] * jc + nodes[k]->v[z] / norm) / (jc + 1);
+	q->v[z] = (q->v[z] * jc + nodes[k]->v[z] / norm) / (jc + 1);
+      Distance::init_node(q, f);
       jc++;
     }
   }
@@ -301,16 +257,20 @@ struct Angular {
      * more memory to be able to fit the vector outside
      */
     S n_descendants;
-    S children[2]; // Will possibly store more than 2
+    union {
+      S children[2]; // Will possibly store more than 2
+      T norm;
+    };
     T v[1]; // We let this one overflow intentionally. Need to allocate at least 1 to make GCC happy
   };
-  template<typename T>
-  static inline T distance(const T* x, const T* y, int f) {
+  template<typename S, typename T>
+  static inline T distance(const Node<S, T>* x, const Node<S, T>* y, int f) {
     // want to calculate (a/|a| - b/|b|)^2
     // = a^2 / a^2 + b^2 / b^2 - 2ab/|a||b|
     // = 2 - 2cos
-    T pp, qq, pq;
-    dot3(x, y, f, &pp, &qq, &pq);
+    T pp = x->norm ? x->norm : dot(x->v, x->v, f); // For backwards compatibility reasons, we need to fall back and compute the norm here
+    T qq = y->norm ? y->norm : dot(y->v, y->v, f);
+    T pq = dot(x->v, y->v, f);
     T ppqq = pp * qq;
     if (ppqq > 0) return 2.0 - 2.0 * pq / sqrt(ppqq);
     else return 2.0; // cos is 0
@@ -328,12 +288,15 @@ struct Angular {
       return random.flip();
   }
   template<typename S, typename T, typename Random>
-  static inline void create_split(const vector<Node<S, T>*>& nodes, int f, Random& random, Node<S, T>* n) {
-    vector<T> best_iv(f, 0), best_jv(f, 0); // TODO: avoid allocation
-    two_means<T, Random, Angular, Node<S, T> >(nodes, f, random, true, &best_iv[0], &best_jv[0]);
+  static inline void create_split(const vector<Node<S, T>*>& nodes, int f, size_t s, Random& random, Node<S, T>* n) {
+    Node<S, T>* p = (Node<S, T>*)malloc(s); // TODO: avoid
+    Node<S, T>* q = (Node<S, T>*)malloc(s); // TODO: avoid
+    two_means<T, Random, Angular, Node<S, T> >(nodes, f, random, true, p, q);
     for (int z = 0; z < f; z++)
-      n->v[z] = best_iv[z] - best_jv[z];
+      n->v[z] = p->v[z] - q->v[z];
     normalize(n->v, f);
+    free(p);
+    free(q);
   }
   template<typename T>
   static inline T normalized_distance(T distance) {
@@ -352,13 +315,16 @@ struct Angular {
   static inline T pq_initial_value() {
     return numeric_limits<T>::infinity();
   }
+  template<typename S, typename T>
+  static inline void init_node(Node<S, T>* n, int f) {
+    n->norm = dot(n->v, n->v, f);
+  }
   static const char* name() {
     return "angular";
   }
 };
 
 struct Hamming {
-
   template<typename S, typename T>
   struct ANNOY_NODE_ATTRIBUTE Node {
     S n_descendants;
@@ -375,13 +341,13 @@ struct Hamming {
 
   template<typename T>
   static inline T pq_initial_value() {
-    return 0;
+    return numeric_limits<T>::max();
   }
-  template<typename T>
-  static inline T distance(const T* x, const T* y, int f) {
+  template<typename S, typename T>
+  static inline T distance(const Node<S, T>* x, const Node<S, T>* y, int f) {
     size_t dist = 0;
     for (int i = 0; i < f; i++) {
-      dist += popcount(x[i] ^ y[i]);
+      dist += popcount(x->v[i] ^ y->v[i]);
     }
     return dist;
   }
@@ -396,12 +362,13 @@ struct Hamming {
     return margin(n, y, f);
   }
   template<typename S, typename T, typename Random>
-  static inline void create_split(const vector<Node<S, T>*>& nodes, int f, Random& random, Node<S, T>* n) {
+  static inline void create_split(const vector<Node<S, T>*>& nodes, int f, size_t s, Random& random, Node<S, T>* n) {
     size_t cur_size = 0;
     size_t i = 0;
+    int dim = f * 8 * sizeof(T);
     for (; i < max_iterations; i++) {
       // choose random position to split at
-      n->v[0] = random.index(f);
+      n->v[0] = random.index(dim);
       cur_size = 0;
       for (typename vector<Node<S, T>*>::const_iterator it = nodes.begin(); it != nodes.end(); ++it) {
         if (margin(n, (*it)->v, f)) {
@@ -415,7 +382,7 @@ struct Hamming {
     // brute-force search for splitting coordinate
     if (i == max_iterations) {
       int j = 0;
-      for (; j < f; j++) {
+      for (; j < dim; j++) {
         n->v[0] = j;
         cur_size = 0;
 	for (typename vector<Node<S, T>*>::const_iterator it = nodes.begin(); it != nodes.end(); ++it) {
@@ -433,6 +400,9 @@ struct Hamming {
   static inline T normalized_distance(T distance) {
     return distance;
   }
+  template<typename S, typename T>
+  static inline void init_node(Node<S, T>* n, int f) {
+  }
   static const char* name() {
     return "hamming";
   }
@@ -443,7 +413,10 @@ struct Minkowski {
   struct ANNOY_NODE_ATTRIBUTE Node {
     S n_descendants;
     T a; // need an extra constant term to determine the offset of the plane
-    S children[2];
+    union {
+      S children[2];
+      T norm;
+    };
     T v[1];
   };
   template<typename S, typename T>
@@ -472,83 +445,68 @@ struct Minkowski {
 
 
 struct Euclidean : Minkowski{
-  template<typename T>
-  static inline T distance(const T* x, const T* y, int f) {
-    T pp, qq, pq;
-    dot3(x, y, f, &pp, &qq, &pq);
+  template<typename S, typename T>
+  static inline T distance(const Node<S, T>* x, const Node<S, T>* y, int f) {
+    T pp = x->norm ? x->norm : dot(x->v, x->v, f); // For backwards compatibility reasons, we need to fall back and compute the norm here
+    T qq = y->norm ? y->norm : dot(y->v, y->v, f);
+    T pq = dot(x->v, y->v, f);
     return pp + qq - 2*pq;
   }
   template<typename S, typename T, typename Random>
-  static inline void create_split(const vector<Node<S, T>*>& nodes, int f, Random& random, Node<S, T>* n) {
-    vector<T> best_iv(f, 0), best_jv(f, 0);
-    two_means<T, Random, Euclidean, Node<S, T> >(nodes, f, random, false, &best_iv[0], &best_jv[0]);
+  static inline void create_split(const vector<Node<S, T>*>& nodes, int f, size_t s, Random& random, Node<S, T>* n) {
+    Node<S, T>* p = (Node<S, T>*)malloc(s); // TODO: avoid
+    Node<S, T>* q = (Node<S, T>*)malloc(s); // TODO: avoid
+    two_means<T, Random, Euclidean, Node<S, T> >(nodes, f, random, false, p, q);
 
     for (int z = 0; z < f; z++)
-      n->v[z] = best_iv[z] - best_jv[z];
+      n->v[z] = p->v[z] - q->v[z];
     normalize(n->v, f);
     n->a = 0.0;
     for (int z = 0; z < f; z++)
-      n->a += -n->v[z] * (best_iv[z] + best_jv[z]) / 2;
+      n->a += -n->v[z] * (p->v[z] + q->v[z]) / 2;
+    free(p);
+    free(q);
   }
   template<typename T>
   static inline T normalized_distance(T distance) {
     return sqrt(std::max(distance, T(0)));
   }
+  template<typename S, typename T>
+  static inline void init_node(Node<S, T>* n, int f) {
+    n->norm = dot(n->v, n->v, f);
+  }
   static const char* name() {
     return "euclidean";
   }
 };
 
 struct Manhattan : Minkowski{
-  template<typename T>
-  static inline T distance(const T* x, const T* y, int f) {
-    T d = 0.0;
-    for (int i = 0; i < f; i++, x++, y++)
-      d += fabs((*x) - (*y));
-    return d;
+  template<typename S, typename T>
+  static inline T distance(const Node<S, T>* x, const Node<S, T>* y, int f) {
+    return manhattan_distance(x->v, y->v, f);
   }
-#ifdef USE_AVX
-  static inline float distance(const float* x, const float* y, int f) {
-    float result = 0;
-    int i = f;
-    if (f > 7) {
-      __m256 manhattan = _mm256_setzero_ps();
-      __m256 minus_zero = _mm256_set1_ps(-0.0f);
-      for (; i > 7; i -= 8) {
-        const __m256 x_minus_y = _mm256_sub_ps(_mm256_loadu_ps(x), _mm256_loadu_ps(y));
-        const __m256 distance = _mm256_andnot_ps(minus_zero, x_minus_y); // Absolute value of x_minus_y (forces sign bit to zero)
-        manhattan = _mm256_add_ps(manhattan, distance);
-        x += 8;
-        y += 8;
-      }
-      // Sum all floats in manhattan register.
-      result = hsum256_ps_avx(manhattan);
-    }
-    // Don't forget the remaining values.
-    for (; i > 0; i--) {
-      result += fabsf(*x - *y);
-      x++;
-      y++;
-    }
-    return result;
-  }
-#endif
   template<typename S, typename T, typename Random>
-  static inline void create_split(const vector<Node<S, T>*>& nodes, int f, Random& random, Node<S, T>* n) {
-    vector<T> best_iv(f, 0), best_jv(f, 0);
-    two_means<T, Random, Manhattan, Node<S, T> >(nodes, f, random, false, &best_iv[0], &best_jv[0]);
+  static inline void create_split(const vector<Node<S, T>*>& nodes, int f, size_t s, Random& random, Node<S, T>* n) {
+    Node<S, T>* p = (Node<S, T>*)malloc(s); // TODO: avoid
+    Node<S, T>* q = (Node<S, T>*)malloc(s); // TODO: avoid
+    two_means<T, Random, Manhattan, Node<S, T> >(nodes, f, random, false, p, q);
 
     for (int z = 0; z < f; z++)
-      n->v[z] = best_iv[z] - best_jv[z];
+      n->v[z] = p->v[z] - q->v[z];
     normalize(n->v, f);
     n->a = 0.0;
     for (int z = 0; z < f; z++)
-      n->a += -n->v[z] * (best_iv[z] + best_jv[z]) / 2;
+      n->a += -n->v[z] * (p->v[z] + q->v[z]) / 2;
+    free(p);
+    free(q);
   }
   template<typename T>
   static inline T normalized_distance(T distance) {
     return std::max(distance, T(0));
   }
+  template<typename S, typename T>
+  static inline void init_node(Node<S, T>* n, int f) {
+  }
   static const char* name() {
     return "manhattan";
   }
@@ -630,6 +588,7 @@ public:
 
     for (int z = 0; z < _f; z++)
       n->v[z] = w[z];
+    D::init_node(n, _f);
 
     if (item >= _n_items)
       _n_items = item + 1;
@@ -655,7 +614,7 @@ public:
           indices.push_back(i);
       }
 
-      _roots.push_back(_make_tree(indices));
+      _roots.push_back(_make_tree(indices, true));
     }
     // Also, copy the roots into the last segment of the array
     // This way we can load them faster without reading the whole file
@@ -752,9 +711,7 @@ public:
   }
 
   T get_distance(S i, S j) {
-    const T* x = _get(i)->v;
-    const T* y = _get(j)->v;
-    return D::normalized_distance(D::distance(x, y, _f));
+    return D::normalized_distance(D::distance(_get(i), _get(j), _f));
   }
 
   void get_nns_by_item(S item, size_t n, size_t search_k, vector<S>* result, vector<T>* distances) {
@@ -798,15 +755,15 @@ protected:
     return (Node*)((uint8_t *)_nodes + (_s * i));
   }
 
-  S _make_tree(const vector<S >& indices) {
-    if (indices.size() == 1)
+  S _make_tree(const vector<S >& indices, bool is_root) {
+    if (indices.size() == 1 && !is_root)
       return indices[0];
 
     if (indices.size() <= (size_t)_K) {
       _allocate_size(_n_nodes + 1);
       S item = _n_nodes++;
       Node* m = _get(item);
-      m->n_descendants = (S)indices.size();
+      m->n_descendants = is_root ? _n_items : (S)indices.size();
 
       // Using std::copy instead of a loop seems to resolve issues #3 and #13,
       // probably because gcc 4.8 goes overboard with optimizations.
@@ -825,7 +782,7 @@ protected:
 
     vector<S> children_indices[2];
     Node* m = (Node*)malloc(_s); // TODO: avoid
-    D::create_split(children, _f, _random, m);
+    D::create_split(children, _f, _s, _random, m);
 
     for (size_t i = 0; i < indices.size(); i++) {
       S j = indices[i];
@@ -857,10 +814,10 @@ protected:
 
     int flip = (children_indices[0].size() > children_indices[1].size());
 
-    m->n_descendants = (S)indices.size();
+    m->n_descendants = is_root ? _n_items : (S)indices.size();
     for (int side = 0; side < 2; side++)
       // run _make_tree for the smallest child first (for cache locality)
-      m->children[side^flip] = _make_tree(children_indices[side^flip]);
+      m->children[side^flip] = _make_tree(children_indices[side^flip], false);
 
     _allocate_size(_n_nodes + 1);
     S item = _n_nodes++;
@@ -871,6 +828,10 @@ protected:
   }
 
   void _get_all_nns(const T* v, size_t n, size_t search_k, vector<S>* result, vector<T>* distances) {
+    Node* v_node = (Node *)malloc(_s); // TODO: avoid
+    memcpy(v_node->v, v, sizeof(T)*_f);
+    D::init_node(v_node, _f);
+
     std::priority_queue<pair<T, S> > q;
 
     if (search_k == (size_t)-1)
@@ -880,7 +841,7 @@ protected:
       q.push(make_pair(Distance::template pq_initial_value<T>(), _roots[i]));
     }
 
-    vector<S> nns;
+    std::vector<S> nns;
     while (nns.size() < search_k && !q.empty()) {
       const pair<T, S>& top = q.top();
       T d = top.first;
@@ -909,7 +870,7 @@ protected:
       if (j == last)
         continue;
       last = j;
-      nns_dist.push_back(make_pair(D::distance(v, _get(j)->v, _f), j));
+      nns_dist.push_back(make_pair(D::distance(v_node, _get(j), _f), j));
     }
 
     size_t m = nns_dist.size();
@@ -920,6 +881,7 @@ protected:
         distances->push_back(D::normalized_distance(nns_dist[i].first));
       result->push_back(nns_dist[i].second);
     }
+    free(v_node);
   }
 };
 
